#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of RFAudioCoder project.
#
# RFAudioCoder is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RFAudioCoder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RFAudioCoder.  If not, see <http://www.gnu.org/licenses/>


# 2015-05-01 David Goncalves <david@lestat.st>
#            RFAudioCoder is a class that enables you to use your audio device to send
#            RF messages with an RF emitter connected on it (315/433MHz emitters).
#
#            The idea was that every device (even very low ones) can send data to an
#            audio port at at least 44100Hz that is almost enough to encode 433MHz
#            RF signals... This class enables you to encode such messages and send
#            them trough audio device or generate WAV files to be played by another
#            machine(s).

import math
import sys
import wave

class RFAudioCoder :

    _debug         = None
    _version       = "1.0.0"
    _audio_freq    = 8000       # Audio encoding frequency
    _msg_repeat    = 8          # Number of repeat times
    _msg_bits      = 24         # Default message length

    # Define a dict containing known protocols
    protocols = { "1" : { "name"        : "Protocol 1",    # Friendly name
                          "bits"        : 24,              # Message length
                          "pulse"       : 350 * 0.000001,  # 350µs
                          "coding_0"    : [ 1, 3 ],        # 1 pulse high and 3 pulses low
                          "coding_1"    : [ 3, 1 ],        # 3 pulses high and 1 pulse low
                          "coding_sync" : [ 1, 31 ]        # 1 pulse high and 31 pulses low
                        },

                  "2" : { "name"        : "Protocol 2",
                          "bits"        : 32,
                          "pulse"       : 650 * 0.000001,
                          "coding_0"    : [ 1, 2 ],
                          "coding_1"    : [ 2, 1 ],
                          "coding_sync" : [ 1, 10 ]
                        }
                }

    def __init__( self, debug=False ) :
        self._debug = debug
        self._debug_message( "Initializing RFAudioCoder class" )


    def _debug_message( self, message="" ) :
        if ( self._debug == True ) :
            print( "[DEBUG] %s" % message )
            sys.stdout.flush()

    def _encode_message( self, protocol="", message=0 ) :
        if ( protocol not in self.protocols.keys() ) :
            raise ValueError( "Protocol '%s' not supported" % protocol )

        self._debug_message( "Encoding message '%s' using '%s'" % ( message, self.protocols[ protocol ]["name"] ) )
        self._debug_message( " -> Repeating message %s times" % self._msg_repeat )
        self._debug_message( " -> Using %s data bits" % self._msg_bits )

        # Determine the audio pulse length
        audio_pulse           = float( 1.0 / self._audio_freq )
        audio_sample_duration = int( math.ceil( self.protocols[ protocol ][ "pulse" ] / audio_pulse ) )

        self._debug_message( " -> Protocol pulse length = %dµs (converted to %d samples @ %dHz)" % ( ( self.protocols[ protocol ][ "pulse" ] * 1000000 ), audio_sample_duration, self._audio_freq ) )

        encoded_buffer = chr(127)  # Init the buffer with a 0V value to avoid voltage peaks at startup

        bit_coding = [ chr( 0 ), chr( 255 ) ]
        #bit_coding = [ chr( 127 ), chr( 255 ) ]  # This bit coding provides positive only pulses

        for repeat in range( self._msg_repeat ) :
            for count in range( 1, self._msg_bits + 1 ) :
                bit = ( message >> ( self._msg_bits - count ) ) & 1

                encoded_buffer += bit_coding[ 1 ] * ( self.protocols[ protocol ][ "coding_%s" % bit ][ 0 ] * audio_sample_duration ) # Encode high states
                encoded_buffer += bit_coding[ 0 ] * ( self.protocols[ protocol ][ "coding_%s" % bit ][ 1 ] * audio_sample_duration ) # Encode low states

            # Add the sync bit
            encoded_buffer += bit_coding[ 1 ] * ( self.protocols[ protocol ][ "coding_sync" ][ 0 ] * audio_sample_duration ) # Encode high states
            encoded_buffer += bit_coding[ 0 ] * ( self.protocols[ protocol ][ "coding_sync" ][ 1 ] * audio_sample_duration ) # Encode low states

        self._debug_message( " -> Encoded message size  = %s bytes" % len( encoded_buffer ) )
        return( encoded_buffer )

    def ExportAudio( self, protocol=None, message=0 ) :
        """Export a message code to default audio device.

        Args:
            protocol: Protocol used to encode message
            message:  Message to encode (numeric format)
        """

        try :
            import alsaaudio

        except ImportError :
            self._debug_message( "Module 'alsaaudio' failed to import, disabling audio device export" )
            return

        data = self._encode_message( protocol, message )

        audiodev = alsaaudio.PCM( type=alsaaudio.PCM_PLAYBACK )
        audiodev.setformat( alsaaudio.PCM_FORMAT_U8 )
        audiodev.setchannels( 1 )
        audiodev.setrate( self._audio_freq )
        audiodev.setperiodsize( 1024 )

        self._debug_message( "Sending message to audio device '%s' (padded to %s bytes)" % ( audiodev.cardname(), len( data ) ) )
        audiodev.write( data )
        audiodev.close()

    def ExportWAV( self, protocol=None, message=0, dest="" ) :
        """Export a message code to a WAV file.

        Args:
            protocol: Protocol used to encode message
            message:  Message to encode (numeric format)
            dest:     Destination WAV file to be written
        """

        data = self._encode_message( protocol, message )
        self._debug_message( "Exporting message to wav file '%s'" % dest )

        # Params = [ nchannels, sampwidth, framerate, nframes, comptype, compname ]
        wav = wave.open( dest, "w" )
        wav.setparams( ( 1, 1, self._audio_freq, len( data ), "NONE", "Uncompressed" ) )
        wav.writeframes( data )
        wav.close()

    def SetAudioFrequency( self, audio_freq=8000 ) :
        """Set the audio frequency used to encode the messages (default: 8000Hz)"""

        self._audio_freq = audio_freq

    def SetMessageRepeat( self, msg_repeat=8 ) :
        """Set the repeat times of the encoded message (default: 8)"""

        self._msg_repeat = msg_repeat

    def SetMessageBitLength( self, msg_bits=24 ) :
        """Set the message data bit length (default: 24)"""
        
        self._msg_bits = msg_bits
